# -*- coding: utf-8 -*-
# pylint: disable-msg = C0103, W0232, W0302, W0612, W0613, W0703, W0704, E0213
# pylint: disable-msg = E1001
#
# Copyright 2004-2006 André Malo or his licensors, as applicable
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Testcases for svnmailer.typedstruct

@var _tests: The test classes
@type _tests: C{list}
"""
__author__    = "André Malo"
__docformat__ = "epytext en"
__all__       = ['getSuite']

import unittest
_tests = []


class importTestCase(unittest.TestCase):
    """ Importing svnmailer.typedstruct """

    def testImport(self):
        """ from svnmailer import typedstruct """
        try:
            from svnmailer import typedstruct
        except Exception, exc:
            self.fail(
                "svnmailer.typedstruct could not be imported: %s" % str(exc)
            )

_tests.append(importTestCase)


class structTestBase(unittest.TestCase):
    """ Base class for struct tests """

    def setUp(self):
        """ set up test environment """
        from svnmailer import typedstruct

        self._struct = typedstruct


class StructPrivateTestCase(structTestBase):
    """ svnmailer.struct.StructPrivate """

    def test00Instance(self):
        """ StructPrivate initializes properly """
        private = self._struct.StructPrivate(['a'], {'b': 'a'})
        expected = ('a',)
        result = private.members
        self.failUnlessEqual(expected, result,
            "members: expected %r, got %r" % (expected, result)
        )

        expected = {'b': 'a'}
        result = private.aliases
        self.failUnlessEqual(expected, result,
            "aliases: expected %r, got %r" % (expected, result)
        )

    def test01Storage(self):
        """ StructPrivate stores data per external instance """
        class C1(object):
            pass
        class C2(object):
            pass
        inst1 = C1()
        inst2 = C2()

        private = self._struct.StructPrivate(['a'], {'b': 'a'})
        private.initInstance(inst1, {'c': 1}, 1)
        private.initInstance(inst2, {'c': 1}, 2)

        val1 = private.getValues(inst1)
        val2 = private.getValues(inst2)
        self.failIf(val1 is val2, "The storage dicts are identical")
        val1['c'] = 3
        self.failIf(val1 is val2, "The storage dicts are identical")

        map1 = private.getMaps(inst1)
        map2 = private.getMaps(inst2)
        self.failIf(map1 is map2, "The map dicts are identical")
        self.failUnless(map1 == {'c': 1} and map2 == {'c': 1},
            "Stored maps are: %r and %r" % (map1, map2)
        )

        arg1 = private.getArg(inst1)
        arg2 = private.getArg(inst2)
        self.failIf(arg1 is arg2, "The args are identical")
        self.failUnless(arg1 == 1 and arg2 == 2,
            "Stored args are: %r and %r" % (arg1, arg2)
        )

        subst1 = private.getSubst(inst1)
        subst2 = private.getSubst(inst2)
        self.failIf(subst1 is subst2, "The subst dicts are identical")
        subst1['a'] = 4
        self.failIf(subst1 is subst2, "The subst dicts are identical")

    def test02aliasedmaps(self):
        """ StructPrivate accepts and merges aliased maps """
        private = self._struct.StructPrivate(['a', 'b'], {'c': 'b'})
        maps = {'a': 1, 'c': 2}
        private.initInstance(self, maps, None)

        expected = {'a': 1, 'c': 2}
        result = maps
        self.failUnlessEqual(expected, result,
            "Input map was modified, expected %r, got %r" % (expected, result)
        )

        expected = {'a': 1, 'b': 2}
        result = private.getMaps(self)
        self.failUnlessEqual(expected, result,
            "Aliases maps were not merged. Expected %r, got %r" %
            (expected, result)
        )

_tests.append(StructPrivateTestCase)


class StructDescriptorTestCase(structTestBase):
    """ svnmailer.struct.StructDescriptor """

    def test00set(self):
        """ StructDescriptor raises AttributeError on __set__ """
        class Described(object):
            x = self._struct.StructDescriptor(None)

        inst = Described()
        try:
            inst.x = 1
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Setting doesn't raise AttributeError")

    def test01delete(self):
        """ StructDescriptor raises AttributeError on __delete__ """
        class Described(object):
            x = self._struct.StructDescriptor(None)

        inst = Described()
        try:
            del inst.x
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Deleting doesn't raise AttributeError")

    def test02inheritance(self):
        """ StructDescriptor inherits its behaviour """
        class MyDescriptor(self._struct.StructDescriptor):
            pass
        class Described(object):
            x = MyDescriptor(None)

        inst = Described()
        try:
            inst.x = 1
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Setting doesn't raise AttributeError")

        try:
            del inst.x
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Deleting doesn't raise AttributeError")

_tests.append(StructDescriptorTestCase)


class StructSetDescriptorTestCase(structTestBase):
    """ svnmailer.struct.StructSetDescriptor """

    def test00base(self):
        """ StructSetDescriptor is read-only """
        class Described(object):
            x = self._struct.StructSetDescriptor(None)

        inst = Described()
        try:
            inst.x = 1
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Setting doesn't raise AttributeError")

        try:
            del inst.x
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Deleting doesn't raise AttributeError")

    def test01setter(self):
        """ StructSetDescriptor.__get__ returns a function reference """
        import types
        private = self._struct.StructPrivate((), {})
        class Described(object):
            x = self._struct.StructSetDescriptor(private)

        inst = Described()
        self.failUnless(isinstance(inst.x, types.FunctionType),
            "Doesn't return a function, but %r" % type(inst.x)
        )

    def test02private(self):
        """ StructSetDescriptor.__get__ stores a private reference """
        private = self._struct.StructPrivate((), {})
        class Described(object):
            x = self._struct.StructSetDescriptor(private)

        inst = Described()
        self.failUnless(inst.x.private is private,
            "Doesn't store a private reference, but %r" % inst.x.private
        )

    def test03set(self):
        """ StructSetDescriptor.__get__ sets values properly """
        private = self._struct.StructPrivate((), {})
        class Described(object):
            _set_ = self._struct.StructSetDescriptor(private)

        inst = Described()
        inst._set_('x', 1)
        self.failUnlessEqual(inst.x, 1, "1st: inst.x is %r" % inst.x)

        inst._set_('y', 2)
        self.failUnlessEqual(inst.x, 1, "2nd: inst.x is %r" % inst.x)
        self.failUnlessEqual(inst.y, 2, "2nd: inst.y is %r" % inst.y)

    def test04alias(self):
        """ StructSetDescriptor.__get__ uses aliases properly """
        private = self._struct.StructPrivate((), {'x': 'a', 'y': 'b'})
        class Described(object):
            _set_ = self._struct.StructSetDescriptor(private)

        inst = Described()
        inst._set_('x', 1)
        self.failUnlessEqual(inst.a, 1, "1st: inst.a is %r" % inst.a)

        inst._set_('y', 2)
        self.failUnlessEqual(inst.a, 1, "2nd: inst.a is %r" % inst.a)
        self.failUnlessEqual(inst.b, 2, "2nd: inst.b is %r" % inst.b)

    def test05choke(self):
        """ StructSetDescriptor.__get__ chokes on private names """
        private = self._struct.StructPrivate((), {})
        class Described(object):
            _set_ = self._struct.StructSetDescriptor(private)

        inst = Described()
        inst._set_('_x', 1)
        inst._set_('x_', 1)
        try:
            inst._set_('_x_', 1)
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Using a private name doesn't raise AttributeError")

_tests.append(StructSetDescriptorTestCase)


class StructSubDescriptorTestCase(structTestBase):
    """ svnmailer.struct.StructSubDescriptor """

    def test00base(self):
        """ StructSubDescriptor is read-only """
        class Described(object):
            x = self._struct.StructSubDescriptor(None)

        inst = Described()
        try:
            inst.x = 1
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Setting doesn't raise AttributeError")

        try:
            del inst.x
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Deleting doesn't raise AttributeError")

    def test01setter(self):
        """ StructSubDescriptor.__get__ returns a function reference """
        import types
        private = self._struct.StructPrivate((), {})
        class Described(object):
            x = self._struct.StructSubDescriptor(private)

        inst = Described()
        self.failUnless(isinstance(inst.x, types.FunctionType),
            "Doesn't return a function, but %r" % type(inst.x)
        )

    def test02set(self):
        """ StructSubDescriptor.__get__ sets subst values properly """
        private = self._struct.StructPrivate((), {})
        class Described(object):
            _sub_ = self._struct.StructSubDescriptor(private)
            def __init__(this):
                private.initInstance(this, None, None)

        inst = Described()
        inst._sub_('x', 1)
        result = private.getSubst(inst)['x']
        self.failUnlessEqual(result, 1, "1st: sub(x) is %r" % result)

        inst._sub_('y', 2)
        result = private.getSubst(inst)['x']
        self.failUnlessEqual(result, 1, "2nd: sub(x) is %r" % result)
        result = private.getSubst(inst)['y']
        self.failUnlessEqual(result, 2, "2nd: sub(y) is %r" % result)

        inst._sub_('x', 3)
        result = private.getSubst(inst)['x']
        self.failUnlessEqual(result, 3, "3rd: sub(x) is %r" % result)
        result = private.getSubst(inst)['y']
        self.failUnlessEqual(result, 2, "3rd: sub(y) is %r" % result)

    def test03set(self):
        """ StructSubDescriptor.__get__ sets subst default values properly """
        private = self._struct.StructPrivate((), {})
        class Described(object):
            _sub_ = self._struct.StructSubDescriptor(private)
            def __init__(this):
                private.initInstance(this, None, None)

        inst = Described()
        inst._sub_('x', 1)
        result = private.getSubst(inst)['x']
        self.failUnlessEqual(result, 1, "1st: sub(x) is %r" % result)

        inst._sub_('x', 2, default = True)
        result = private.getSubst(inst)['x']
        self.failUnlessEqual(result, 1, "2nd: sub(x) is %r" % result)

    def test04set(self):
        """ StructSubDescriptor.__get__ provides the current dict """
        private = self._struct.StructPrivate((), {})
        class Described(object):
            _sub_ = self._struct.StructSubDescriptor(private)
            def __init__(this):
                private.initInstance(this, None, None)

        inst = Described()
        inst._sub_('x', 1)
        inst._sub_('y', 2)
        expected = {'x': 1, 'y': 2}
        result = inst._sub_.dict()
        self.failUnlessEqual(expected, result,
            "Expected %r, got %r" % (expected, result)
        )

    def test05choke(self):
        """ StructSubDescriptor.__get__ chokes on class access """
        private = self._struct.StructPrivate((), {})
        class Described(object):
            _sub_ = self._struct.StructSubDescriptor(private)
            def __init__(this):
                private.initInstance(this, None, None)

        try:
            Described._sub_('x', 1)
            self.fail("Class access doesn't raise AttributeError")
        except AttributeError:
            # cool
            pass

_tests.append(StructSubDescriptorTestCase)


class StructMembersDescriptorTestCase(structTestBase):
    """ svnmailer.struct.StructMembersDescriptor """

    def test00base(self):
        """ StructMembersDescriptor is read-only """
        class Described(object):
            x = self._struct.StructMembersDescriptor(None)

        inst = Described()
        try:
            inst.x = 1
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Setting doesn't raise AttributeError")

        try:
            del inst.x
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Deleting doesn't raise AttributeError")

    def test01return(self):
        """ StructMembersDescriptor returns a proper tuple """
        private = self._struct.StructPrivate(('a', 'b', 'f'), {})
        class Described(object):
            _members_ = self._struct.StructMembersDescriptor(private)
            def __init__(this):
                private.initInstance(this, None, None)

        expected = ('a', 'b', 'f')
        result = Described._members_
        self.failUnlessEqual(expected, result,
            "Class access: expected %r, got %r" % (expected, result)
        )
        result = Described()._members_
        self.failUnlessEqual(expected, result,
            "Instance access: expected %r, got %r" % (expected, result)
        )

_tests.append(StructMembersDescriptorTestCase)


class StructDictDescriptorTestCase(structTestBase):
    """ svnmailer.struct.StructDictDescriptor """

    def test00base(self):
        """ StructDictDescriptor is read-only """
        class Described(object):
            x = self._struct.StructDictDescriptor(None)

        inst = Described()
        try:
            inst.x = 1
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Setting doesn't raise AttributeError")

        try:
            del inst.x
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Deleting doesn't raise AttributeError")

    def test01return(self):
        """ StructDictDescriptor returns a proper dict """
        private = self._struct.StructPrivate(('a', 'b', 'f'), {})
        class Described(object):
            _dict_ = self._struct.StructDictDescriptor(private)
            def __init__(this):
                private.initInstance(this, None, None)

        inst = Described()
        private.getValues(inst).update({'a': 1, 'b': 2, 'f': 3})

        expected = {'a': None, 'b': None, 'f': None}
        result = Described._dict_
        self.failUnlessEqual(expected, result,
            "Class access: expected %r, got %r" % (expected, result)
        )

        expected = {'a': 1, 'b': 2, 'f': 3}
        result = inst._dict_
        self.failUnlessEqual(expected, result,
            "Instance access: expected %r, got %r" % (expected, result)
        )

_tests.append(StructDictDescriptorTestCase)


class MemberDescriptorTestCase(structTestBase):
    """ svnmailer.struct.MemberDescriptor """

    def test00public(self):
        """ MemberDescriptor's name and param are public attributes """
        priv = self._struct.StructPrivate((), {})
        desc = self._struct.MemberDescriptor('myname', priv, self)

        expected = ('myname', self)
        result = (desc.name, desc.param)
        self.failUnlessEqual(expected, result,
            "Expected %r, got %r" % (expected, result)
        )

    def test01delete(self):
        """ MemberDescriptor.__delete__ raises AttributeError """
        priv = self._struct.StructPrivate(('myname',), {})
        desc = self._struct.MemberDescriptor('myname', priv, self)
        priv.initInstance(self, None, None)

        try:
            desc.__delete__(self)
        except AttributeError:
            # right
            pass
        else:
            self.fail("__delete__ doesn't raise an AttributeError")

    def test02set(self):
        """ MemberDescriptor.__set__ stores the value """
        priv = self._struct.StructPrivate(('myname',), {})
        desc = self._struct.MemberDescriptor('myname', priv, self)
        priv.initInstance(self, None, None)

        desc.__set__(self, "foo")
        expected = "foo"
        result = priv.getValues(self).get("myname")
        self.failUnlessEqual(expected, result,
            "Expected %r, got %r" % (expected, result)
        )

    def test03set(self):
        """ MemberDescriptor.__set__ premaps the value """
        priv = self._struct.StructPrivate(('myname',), {})

        class MyDescriptor(self._struct.MemberDescriptor):
            def premap(this, value, mapper, arg):
                return (value, "premap", arg, mapper)

        desc = MyDescriptor('myname', priv, self)
        priv.initInstance(self, None, "bar")

        desc.__set__(self, "foo")
        expected = "foo"
        result = priv.getValues(self).get("myname")
        self.failUnlessEqual(expected, result,
            "1st: Expected %r, got %r" % (expected, result)
        )

        priv.initInstance(self, {"myname": 1}, "baz")
        desc.__set__(self, "zonk")
        expected = ("zonk", "premap", "baz", 1)
        result = priv.getValues(self).get("myname")
        self.failUnlessEqual(expected, result,
            "2nd: Expected %r, got %r" % (expected, result)
        )

    def test04set(self):
        """ MemberDescriptor.__set__ transforms the value """
        priv = self._struct.StructPrivate(('myname',), {})

        class MyDescriptor(self._struct.MemberDescriptor):
            def transform(this, value, arg):
                return (value, "transform", arg)

        desc = MyDescriptor('myname', priv, self)
        priv.initInstance(self, None, "bar")

        desc.__set__(self, "foo")
        expected = ("foo", "transform", "bar")
        result = priv.getValues(self).get("myname")
        self.failUnlessEqual(expected, result,
            "Expected %r, got %r" % (expected, result)
        )

    def test05set(self):
        """ MemberDescriptor.__set__ premaps/transforms in order """
        priv = self._struct.StructPrivate(('myname',), {})

        class MyDescriptor(self._struct.MemberDescriptor):
            def premap(this, value, mapper, arg):
                return (value, "premap", arg, mapper)
            def transform(this, value, arg):
                return (value, "transform", arg)

        desc = MyDescriptor('myname', priv, self)
        priv.initInstance(self, {"myname": 1}, "baz")
        desc.__set__(self, "zonk")

        expected = (("zonk", "premap", "baz", 1), "transform", "baz")
        result = priv.getValues(self).get("myname")
        self.failUnlessEqual(expected, result,
            "Expected %r, got %r" % (expected, result)
        )

    def test06get(self):
        """ MemberDescriptor.__get__ returns None on class access """
        priv = self._struct.StructPrivate(('myname',), {})
        desc = self._struct.MemberDescriptor('myname', priv, self)

        expected = None
        result = desc.__get__(None, self.__class__)
        self.failUnlessEqual(expected, result,
            "Expected %r, got %r" % (expected, result)
        )

    def test07get(self):
        """ MemberDescriptor.__get__ returns the stored value """
        priv = self._struct.StructPrivate(('myname',), {})
        desc = self._struct.MemberDescriptor('myname', priv, self)
        priv.initInstance(self, None, None)

        priv.getValues(self)["myname"] = "foo"
        expected = "foo"
        result = desc.__get__(self, self.__class__)
        self.failUnlessEqual(expected, result,
            "Expected %r, got %r" % (expected, result)
        )

    def test08get(self):
        """ MemberDescriptor.__get__ substitutes the value """
        priv = self._struct.StructPrivate(('myname',), {})

        class MyDescriptor(self._struct.MemberDescriptor):
            def substitute(this, value, subst, arg):
                return (value, "substitute", arg, subst)

        desc = MyDescriptor('myname', priv, self)
        priv.initInstance(self, None, "bar")
        priv.getSubst(self)['x'] = 'y'

        priv.getValues(self)["myname"] = "foo"
        expected = ("foo", "substitute", "bar", {"x": "y"})
        result = desc.__get__(self, self.__class__)
        self.failUnlessEqual(expected, result,
            "Expected %r, got %r" % (expected, result)
        )

    def test09get(self):
        """ MemberDescriptor.__get__ postmaps the value """
        priv = self._struct.StructPrivate(('myname',), {})

        class MyDescriptor(self._struct.MemberDescriptor):
            def postmap(this, value, mapper, arg):
                return (value, "postmap", arg, mapper)

        desc = MyDescriptor('myname', priv, self)
        priv.initInstance(self, None, "bar")

        priv.getValues(self)["myname"] = "foo"
        expected = "foo"
        result = desc.__get__(self, self.__class__)
        self.failUnlessEqual(expected, result,
            "1st: Expected %r, got %r" % (expected, result)
        )

        priv.initInstance(self, {"myname": 1}, "baz")
        priv.getValues(self)["myname"] = "zonk"
        expected = ("zonk", "postmap", "baz", 1)
        result = desc.__get__(self, self.__class__)
        self.failUnlessEqual(expected, result,
            "2nd: Expected %r, got %r" % (expected, result)
        )

    def test10get(self):
        """ MemberDescriptor.__get__ substitutes/postmaps in order """
        priv = self._struct.StructPrivate(('myname',), {})

        class MyDescriptor(self._struct.MemberDescriptor):
            def substitute(this, value, subst, arg):
                return (value, "substitute", arg, subst)
            def postmap(this, value, mapper, arg):
                return (value, "postmap", arg, mapper)

        desc = MyDescriptor('myname', priv, self)
        priv.initInstance(self, {"myname": 1}, "baz")
        priv.getSubst(self)['x'] = 'y'

        priv.getValues(self)["myname"] = "zonk"
        expected = (
            ("zonk", "substitute", "baz", {"x": "y"}), "postmap", "baz", 1
        )
        result = desc.__get__(self, self.__class__)
        self.failUnlessEqual(expected, result,
            "Expected %r, got %r" % (expected, result)
        )

    def test11descriptor(self):
        """ MemberDescriptor is accepted as member descriptor """
        priv = self._struct.StructPrivate(('myname',), {})

        class Described(object):
            myname = self._struct.MemberDescriptor('myname', priv, None)

        expected = None
        result = Described.myname
        self.failUnlessEqual(expected, result,
            "Expected %r, got %r" % (expected, result)
        )

        inst = Described()
        priv.initInstance(inst, {"myname": 1}, "foo")
        inst.myname = "bar"

        expected = "bar"
        result = priv.getValues(inst).get("myname")
        self.failUnlessEqual(expected, result,
            "setter: Expected %r, got %r" % (expected, result)
        )

        priv.getValues(inst)["myname"] = "baz"
        expected = "baz"
        result = inst.myname
        self.failUnlessEqual(expected, result,
            "getter: Expected %r, got %r" % (expected, result)
        )

        try:
            del inst.myname
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Member deletion doesn't raise AttributeError")

    def test12slots(self):
        """ MemberDescriptor as descriptor doesn't interfer with __slots__ """
        priv = self._struct.StructPrivate(('myname',), {})

        class Described(object):
            myname = self._struct.MemberDescriptor('myname', priv, None)
            __slots__ = ['myname']

        expected = None
        result = Described.myname
        self.failUnlessEqual(expected, result,
            "Expected %r, got %r" % (expected, result)
        )

        inst = Described()
        priv.initInstance(inst, {"myname": 1}, "foo")
        inst.myname = "bar"

        expected = "bar"
        result = priv.getValues(inst).get("myname")
        self.failUnlessEqual(expected, result,
            "setter: Expected %r, got %r" % (expected, result)
        )

        priv.getValues(inst)["myname"] = "baz"
        expected = "baz"
        result = inst.myname
        self.failUnlessEqual(expected, result,
            "getter: Expected %r, got %r" % (expected, result)
        )

        try:
            del inst.myname
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Member deletion doesn't raise AttributeError")

        try:
            inst.foo = 1
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Unknown member could be set")

    def test13inheriteddescriptor(self):
        """ MemberDescriptor subclass is accepted as member descriptor """
        priv = self._struct.StructPrivate(('myname',), {})

        class MyDescriptor(self._struct.MemberDescriptor):
            def premap(this, value, mapper, arg):
                return (value, "premap")
            def transform(this, value, arg):
                return (value, "transform")
            def substitute(this, value, subst, arg):
                return (value, "substitute")
            def postmap(this, value, mapper, arg):
                return (value, "postmap")

        class Described(object):
            myname = MyDescriptor('myname', priv, None)

        expected = None
        result = Described.myname
        self.failUnlessEqual(expected, result,
            "Expected %r, got %r" % (expected, result)
        )

        inst = Described()
        priv.initInstance(inst, {"myname": 1}, "foo")
        inst.myname = "bar"

        expected = (("bar", "premap"), "transform")
        result = priv.getValues(inst).get("myname")
        self.failUnlessEqual(expected, result,
            "setter: Expected %r, got %r" % (expected, result)
        )

        priv.getValues(inst)["myname"] = "baz"
        expected = (("baz", "substitute"), "postmap")
        result = inst.myname
        self.failUnlessEqual(expected, result,
            "getter: Expected %r, got %r" % (expected, result)
        )

        try:
            del inst.myname
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Member deletion doesn't raise AttributeError")

        inst.myname = "zonk"
        expected = (
            ((("zonk", "premap"), "transform"), "substitute"), "postmap"
        )
        result = inst.myname
        self.failUnlessEqual(expected, result,
            "set/get cycle: Expected %r, got %r" % (expected, result)
        )

_tests.append(MemberDescriptorTestCase)


class membersTestCase(structTestBase):
    """ svnmailer.struct.members """

    def test00members(self):
        """ members accepts the members as list/tuple/dict """
        tests = (('a', 'b'), ['a', 'b'], {'a': None, 'b': None})
        expected = ['a', 'b']

        for test in tests:
            result = self._struct.members({}, test)
            self.failUnlessEqual(expected, result,
                "input %r expected %r, got %r" % (test, expected, result)
            )

    def test01invalidnames(self):
        """ members chokes on invalid members names """
        try:
            self._struct.members({}, ['_a_'])
        except AttributeError:
            # cool
            pass
        else:
            self.fail("Doesn't raise AttributeError on invalid name")

    def test02spacemembers(self):
        """ members exposes the members in supplied space """
        tests = (('a', 'b'), ['a', 'b'], {'a': None, 'b': None})
        expected = ('a', 'b')

        for test in tests:
            space = {}
            self._struct.members(space, test)
            self.failUnless(
                space.has_key('a') and space.has_key('b'),
                "input %r doesn't expose %r to space" % (test, expected)
            )

    def test03nonetypes(self):
        """ members uses MemberDescriptor for None types """
        tests = (('a', 'b'), ['a', 'b'], {'a': None, 'b': None})
        expected = self._struct.MemberDescriptor

        for test in tests:
            space = {}
            self._struct.members(space, test)
            result = (space['a'], space['b'])
            self.failUnless(
                isinstance(result[0], expected) and
                isinstance(result[1], expected),
                "input %r expected %r, got %r" % (test, expected, result)
            )

    def test04simpletypes(self):
        """ members maps simple types from supplied typemap """
        class Type1(self._struct.MemberDescriptor):
            pass
        class Type2(self._struct.MemberDescriptor):
            pass
        typemap = {'1': Type1, 2: Type2}
        members = {'mem1': '1', 'mem2': 2}

        expected = (Type1, Type2)
        space = {}
        self._struct.members(space, members, typemap = typemap)
        result = (space['mem1'], space['mem2'])
        self.failUnless(
            isinstance(result[0], Type1) and isinstance(result[1], Type2),
            "Expected %r, got %r" % (expected, result)
        )

    def test05complextypes(self):
        """ members maps complex types from supplied typemap """
        class Type1(self._struct.MemberDescriptor):
            pass
        class Type2(self._struct.MemberDescriptor):
            pass
        typemap = {'1': Type1, 2: Type2}
        members = {'mem1': ('1', "foo"), 'mem2': 2}

        expected = (Type1, Type2)
        space = {}
        self._struct.members(space, members, typemap = typemap)
        result = (space['mem1'], space['mem2'])
        self.failUnless(
            isinstance(result[0], Type1) and isinstance(result[1], Type2),
            "types: Expected %r, got %r" % (expected, result)
        )

        expected = ("foo", None)
        result = (space['mem1'].param, space['mem2'].param)
        self.failUnlessEqual(expected, result,
            "param: expected %r, got %r" % (expected, result)
        )

    def test06spacespecials(self):
        """ members exposes special attributes in space """
        expected = {
            '_set_'    : self._struct.StructSetDescriptor,
            '_sub_'    : self._struct.StructSubDescriptor,
            '_members_': self._struct.StructMembersDescriptor,
            '_dict_'   : self._struct.StructDictDescriptor,
        }
        space = {}
        self._struct.members(space, ['a'])
        result = dict([(key, space.get(key))
            for key in ['_set_', '_sub_', '_members_', '_dict_']
        ])
        for key, value in expected.items():
            self.failUnless(
                isinstance(result[key], value),
                "Expected %r, got %r" % (expected, result)
            )

    def test07aliases(self):
        """ members stores aliases for later use """
        expected = {'b': 'a'}
        space = {}
        self._struct.members(space, ['a'], aliases = {'b': 'a'})
        setter = space['_set_'].__get__(None, self.__class__)
        private = setter.private
        result = private.aliases

        self.failUnlessEqual(expected, result,
            "Expected %r, got %r" % (expected, result)
        )

_tests.append(membersTestCase)


class StructTestCase(structTestBase):
    """ svnmailer.struct.Struct """

    def test00new(self):
        """ Struct instanciating initializes the private container """
        class MyStruct(self._struct.Struct):
            __slots__ = self._struct.members(locals(), ["a"])

        inst = MyStruct(_maps_ = {"a": self}, _arg_ = "foo")
        inst._set_.private.getValues(inst)

        expected = ({'a': self}, "foo")
        result = (
            inst._set_.private.getMaps(inst),
            inst._set_.private.getArg(inst)
        )
        self.failUnlessEqual(expected, result,
            "expected %r, got %r" % (expected, result)
        )

    def test01init(self):
        """ Struct instanciating assigns names and aliases """
        class MyStruct(self._struct.Struct):
            __slots__ = self._struct.members(locals(), ["a", "b"], {"c": "b"})

        inst = MyStruct(a = 1, c = 2)

        expected = (1, 2)
        result = (inst.a, inst.b)
        self.failUnlessEqual(expected, result,
            "expected %r, got %r" % (expected, result)
        )

    def test02del(self):
        """ Struct destruction cleans up the private container """
        class MyStruct(self._struct.Struct):
            __slots__ = self._struct.members(locals(), ["a"])

        inst = MyStruct()
        private = inst._set_.private
        del inst

        expected = ({}, {}, {})
        result = (private.values, private.maps, private.args)
        self.failUnlessEqual(expected, result,
            "expected %r, got %r" % (expected, result)
        )

    def test03repr(self):
        """ Struct represents itself properly """
        class MyStruct(self._struct.Struct):
            __slots__ = self._struct.members(locals(), ["a"])

        inst = MyStruct(a = 10)

        expected = (
            "%s.MyStruct(\n    _maps_ = {},\n    _arg_ = None,\n    a = 10\n)" %
            self.__class__.__module__
        )
        result = repr(inst)
        self.failUnlessEqual(expected, result,
            "expected %r, got %r" % (expected, result)
        )

    def test04sub(self):
        """ Struct stores substitution pairs properly """
        class MyStruct(self._struct.Struct):
            __slots__ = self._struct.members(locals(), ["a"])

        inst = MyStruct()
        inst._sub_('mykey', 'myvalue')

        expected = {'mykey': 'myvalue'}
        result = inst._set_.private.getSubst(inst)
        self.failUnlessEqual(expected, result,
            "expected %r, got %r" % (expected, result)
        )

    def test05members(self):
        """ Struct allows access to the members list """
        class MyStruct(self._struct.Struct):
            __slots__ = self._struct.members(locals(), ["x", "y"])

        inst = MyStruct()
        expected = ["x", "y"]
        result = list(inst._members_)
        result.sort()
        self.failUnlessEqual(expected, result,
            "expected %r, got %r" % (expected, result)
        )

    def test06dict(self):
        """ Struct allows access to the value dict """
        class MyStruct(self._struct.Struct):
            __slots__ = self._struct.members(locals(), ["a", "b"])

        inst = MyStruct(a = 10, b = 20)
        expected = {'a': 10, 'b': 20}
        result = inst._dict_
        self.failUnlessEqual(expected, result,
            "expected %r, got %r" % (expected, result)
        )

    def test07instances(self):
        """ Struct doesn't confuse different instance values """
        class MyStruct(self._struct.Struct):
            __slots__ = self._struct.members(locals(), ["a"])

        inst1 = MyStruct(a = 10,    _maps_ = {'a': 1}, _arg_ = 3)
        inst2 = MyStruct(a = "foo", _maps_ = {'a': 2}, _arg_ = 4)

        expected = (
            "%s.MyStruct(\n    _maps_ = {'a': 1},\n    _arg_ = 3,\n    "
            "a = 10\n)" % self.__class__.__module__,
            "%s.MyStruct(\n    _maps_ = {'a': 2},\n    _arg_ = 4,\n    "
            "a = 'foo'\n)" % self.__class__.__module__,
        )
        result = (repr(inst1), repr(inst2))
        self.failUnlessEqual(expected, result,
            "expected %r, got %r" % (expected, result)
        )

    def test08instances(self):
        """ Struct doesn't confuse different instance subst pairs """
        class MyStruct(self._struct.Struct):
            __slots__ = self._struct.members(locals(), ["a"])

        inst1 = MyStruct()
        inst2 = MyStruct()
        inst1._sub_("key1", "value1")
        inst2._sub_("key2", "value2")

        expected = ({"key1": "value1"}, {"key2": "value2"})
        result = (
            inst1._set_.private.getSubst(inst1),
            inst2._set_.private.getSubst(inst2)
        )
        self.failUnlessEqual(expected, result,
            "expected %r, got %r" % (expected, result)
        )

_tests.append(StructTestCase)


#### Glue code                                      ####
########################################################
__all__.extend(_tests)

def getSuite():
    """ Returns the test suite for this module """
    return unittest.TestSuite([
        unittest.makeSuite(test) for test in _tests
    ])

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity = 2).run(getSuite())
